/**************************************************************************
ETHERBOOT -  BOOTP/TFTP Bootstrap Program

Author: Martin Renters
  Date: May/94

 This code is based heavily on David Greenman's if_ed.c driver

 Copyright (C) 1993-1994, David Greenman, Martin Renters.
  This software may be used, modified, copied, distributed, and sold, in
  both source and binary form provided that the above copyright and these
  terms are retained. Under no circumstances are the authors responsible for
  the proper functioning of this software, nor do the authors assume any
  responsibility for damages incurred with its use.

3c503 support added by Bill Paul (wpaul@ctr.columbia.edu) on 11/15/94
SMC8416 support added by Bill Paul (wpaul@ctr.columbia.edu) on 12/25/94
3c503 PIO support added by Jim Hague (jim.hague@acm.org) on 2/17/98
RX overrun by Klaus Espenlaub (espenlaub@informatik.uni-ulm.de) on 3/10/99
  parts taken from the Linux 8390 driver (by Donald Becker and Paul Gortmaker)

NE2000 Ethernet Driver for uip
pyrasis (at) chol.com   2003.11.3
**************************************************************************/

#include "ne2000.h"
#include "uip.h"
#include <dos.h>

#include "uip.h"
#include "uipopt.h"

#define inb(a) inp(a)             
#define inw(a) inpw(a)
#define outb(d, a) outp(a, d)         /* gcc outb, dos outp function */
#define outw(d, a) outpw(a, d)        /* gcc outw, dos outpw function */

#define BUF ((struct uip_eth_hdr *)&uip_buf[0])

typedef	unsigned long Address;

#define ETH_ALEN		6	/* Size of Ethernet address */
#define ETH_HLEN		14	/* Size of ethernet header */
#define	ETH_ZLEN		60	/* Minimum packet */
/* #define ETH_MIN_PACKET		64*/
#define	ETH_FRAME_LEN		1514	/* Maximum packet */
/* #define ETH_MAX_PACKET		1518*/
#define	ETH_MAX_MTU		(ETH_FRAME_LEN - ETH_HLEN)

#define	ASIC_PIO NE_DATA

static unsigned char eth_drain_receiver, eth_flags, eth_vendor;
static unsigned short eth_nic_base = 0x300, eth_asic_base;
static Address eth_bmem, eth_rmem;

static unsigned char eth_memsize, eth_tx_start, eth_rx_start;

unsigned short *probe_addrs;

static void eth_pio_write(const unsigned char *src, unsigned int dst, unsigned int cnt)
{

	if (eth_flags & FLAG_16BIT) { ++cnt; cnt &= ~1; }
	outb(D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base + D8390_P0_COMMAND);
	outb(D8390_ISR_RDC, eth_nic_base + D8390_P0_ISR);
	outb(cnt, eth_nic_base + D8390_P0_RBCR0);
	outb(cnt>>8, eth_nic_base + D8390_P0_RBCR1);
	outb(dst, eth_nic_base + D8390_P0_RSAR0);
	outb(dst>>8, eth_nic_base + D8390_P0_RSAR1);
	outb(D8390_COMMAND_RD1 | D8390_COMMAND_STA, eth_nic_base + D8390_P0_COMMAND);
	if (eth_flags & FLAG_16BIT)
		cnt >>= 1;

	while(cnt--)
	{
		if (eth_flags & FLAG_16BIT) {
			outw(*((unsigned short *)src), eth_asic_base + ASIC_PIO);
			src += 2;
		}
		else
			outb(*(src++), eth_asic_base + ASIC_PIO);
	}
}

static void eth_pio_read(unsigned int src, unsigned char *dst, unsigned int cnt)
{
	if (eth_flags & FLAG_16BIT) { ++cnt; cnt &= ~1; }
	outb(D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base + D8390_P0_COMMAND);
	outb(cnt, eth_nic_base + D8390_P0_RBCR0);
	outb(cnt>>8, eth_nic_base + D8390_P0_RBCR1);
	outb(src, eth_nic_base + D8390_P0_RSAR0);
	outb(src>>8, eth_nic_base + D8390_P0_RSAR1);
	outb(D8390_COMMAND_RD0 | D8390_COMMAND_STA, eth_nic_base + D8390_P0_COMMAND);

	if (eth_flags & FLAG_16BIT)
		cnt >>= 1;

	while(cnt--) {
		if (eth_flags & FLAG_16BIT) {
			*((unsigned short *)dst) = inw(eth_asic_base + ASIC_PIO);
			dst += 2;
		}
		else
			*(dst++) = inb(eth_asic_base + ASIC_PIO);
	}
}

int ne2000_probe(struct nic *nic)
{
	int i;
    /*
	struct wd_board *brd;
	unsigned short chksum;
    */
	unsigned char c;
	eth_vendor = VENDOR_NONE;
	eth_drain_receiver = 0;



	if (eth_vendor == VENDOR_NONE) {
		char romdata[16], testbuf[32];
		/*
		int idx;
		static char test[] = "NE*000 memory";
		*/
		static unsigned short base[] = {
/*
#ifdef	NE_SCAN
		NE_SCAN,
#endif
*/
			0 };
		/* if no addresses supplied, fall back on defaults */
		/*
		if (probe_addrs == 0 || probe_addrs[0] == 0)
			probe_addrs = base;
		eth_bmem = 0;	*/	/* No shared memory */
		/* for (idx = 0; (eth_nic_base = probe_addrs[idx]) != 0; ++idx) { */
		eth_nic_base = 0x300;
			eth_flags = FLAG_PIO;
			eth_asic_base = eth_nic_base + NE_ASIC_OFFSET;
			eth_memsize = MEM_16384;
			eth_tx_start = 32;
			eth_rx_start = 32 + D8390_TXBUF_SIZE;
			c = inb(eth_asic_base + NE_RESET);
			outb(c, eth_asic_base + NE_RESET);
			inb(0x84);
			outb(D8390_COMMAND_STP | D8390_COMMAND_RD2, eth_nic_base + D8390_P0_COMMAND);
			outb(D8390_RCR_MON, eth_nic_base + D8390_P0_RCR);
			outb(D8390_DCR_FT1 | D8390_DCR_LS, eth_nic_base + D8390_P0_DCR);
			outb(MEM_8192, eth_nic_base + D8390_P0_PSTART);
			outb(MEM_16384, eth_nic_base + D8390_P0_PSTOP);
#if 0
#ifdef	NS8390_FORCE_16BIT
		eth_flags |= FLAG_16BIT;	/* force 16-bit mode */
#endif
#endif

			/* -eth_pio_write(test, 8192, sizeof(test)); */
			/* -eth_pio_read(8192, testbuf, sizeof(test)); */
			/*
			if (!memcmp(test, testbuf, sizeof(test)))
				break;
			*/
			eth_flags |= FLAG_16BIT;
			eth_memsize = MEM_32768;
			eth_tx_start = 64;
			eth_rx_start = 64 + D8390_TXBUF_SIZE;
			outb(D8390_DCR_WTS | D8390_DCR_FT1 | D8390_DCR_LS, eth_nic_base + D8390_P0_DCR);
			outb(MEM_16384, eth_nic_base + D8390_P0_PSTART);
			outb(MEM_32768, eth_nic_base + D8390_P0_PSTOP);
			/* -eth_pio_write(test, 16384, sizeof(test)); */
			/* -eth_pio_read(16384, testbuf, sizeof(test)); */
			/*
			if (!memcmp(testbuf, test, sizeof(test)))
				break;
			*/
		
		if (eth_nic_base == 0)
			return (0);
		if (eth_nic_base > ISA_MAX_ADDR)	/* PCI probably */
			eth_flags |= FLAG_16BIT;
		eth_vendor = VENDOR_NOVELL;
		eth_pio_read(0, romdata, sizeof(romdata));
		for (i=0; i<ETH_ALEN; i++) {
			nic->addr[i] = romdata[i + ((eth_flags & FLAG_16BIT) ? i : 0)];
		}
		/* printf("\nNE%c000 base %#hx, addr %!\n", (eth_flags & FLAG_16BIT) ? '2' : '1', eth_nic_base, nic->node_addr); */
	}
	if (eth_vendor == VENDOR_NONE)
		return(0);
	/*
	if (eth_vendor != VENDOR_3COM)
		eth_rmem = eth_bmem;

	ne2000_reset(nic);
	nic->reset = ns8390_reset;
	nic->poll = ns8390_poll;
	nic->transmit = ns8390_transmit;
	nic->disable = ns8390_disable;

	return(nic);
	*/
}

void ne2000_reset(struct nic *nic)
{
	int i;

	eth_drain_receiver = 0;

	outb(D8390_COMMAND_PS0 | D8390_COMMAND_RD2 | D8390_COMMAND_STP, eth_nic_base+D8390_P0_COMMAND);
	if (eth_flags & FLAG_16BIT)
		outb(0x49, eth_nic_base+D8390_P0_DCR);
	else
		outb(0x48, eth_nic_base+D8390_P0_DCR);
	outb(0, eth_nic_base+D8390_P0_RBCR0);
	outb(0, eth_nic_base+D8390_P0_RBCR1);
	outb(0x20, eth_nic_base+D8390_P0_RCR);	/* monitor mode */
	outb(2, eth_nic_base+D8390_P0_TCR);
	outb(eth_tx_start, eth_nic_base+D8390_P0_TPSR);
	outb(eth_rx_start, eth_nic_base+D8390_P0_PSTART);

	outb(eth_memsize, eth_nic_base+D8390_P0_PSTOP);
	outb(eth_memsize - 1, eth_nic_base+D8390_P0_BOUND);
	outb(0xFF, eth_nic_base+D8390_P0_ISR);
	outb(0, eth_nic_base+D8390_P0_IMR);

	outb(D8390_COMMAND_PS1 | D8390_COMMAND_RD2 | D8390_COMMAND_STP, eth_nic_base+D8390_P0_COMMAND);
	for (i=0; i<ETH_ALEN; i++)
		outb(nic->addr[i], eth_nic_base+D8390_P1_PAR0+i);
	for (i=0; i<ETH_ALEN; i++)
		outb(0xFF, eth_nic_base+D8390_P1_MAR0+i);
	outb(eth_rx_start, eth_nic_base+D8390_P1_CURR);

	outb(D8390_COMMAND_PS0 | D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base+D8390_P0_COMMAND);
	outb(0xFF, eth_nic_base+D8390_P0_ISR);
	outb(0, eth_nic_base+D8390_P0_TCR);
	outb(4, eth_nic_base+D8390_P0_RCR);	/* allow broadcast frames */
}

#if 0
/* Ignore here */
void ne2000_transmit(
	struct nic *nic,
	const char *d,			/* Destination */
	unsigned int t,			/* Type */
	unsigned int s,			/* size */
	const char *p)			/* Packet */
{
#endif

#if 0
/* Ignore here */
void ne2000_transmit(struct nic *nic, unsigned short t, const char *d)
{
	char tmpbuf[UIP_BUFSIZE];
	int i;
	unsigned short type;

  /*
   * UIP_LLH_LEN 14 : ũ   
   * uipopt.h
   *
   * extern u8_t uip_buf[UIP_BUFSIZE+2];
   * uip.h
   */
	for(i = 0; i < 40 + UIP_LLH_LEN; i++) {
		tmpbuf[i] = uip_buf[i];
	}
  
	for(; i < uip_len; i++) {
		tmpbuf[i] = uip_appdata[i - 40 - UIP_LLH_LEN];
	}

	/* Programmed I/O */
	type = (t >> 8) | (t << 8);
	type = BUF->type;
	type = t;

	eth_pio_write(d, eth_tx_start<<8, ETH_ALEN);
	eth_pio_write(BUF->dest->addr, eth_tx_start<<8, ETH_ALEN);
	
  
	eth_pio_write(nic->addr, (eth_tx_start<<8)+ETH_ALEN, ETH_ALEN);
	/* bcc generates worse code without (const+const) below */
	
  
	eth_pio_write((unsigned char *)&type, (eth_tx_start<<8)+(ETH_ALEN+ETH_ALEN), 2);
	
	eth_pio_write(p, (eth_tx_start<<8)+ETH_HLEN, s);
	eth_pio_write(tmpbuf, (eth_tx_start<<8)+ETH_HLEN, uip_len);
	s += ETH_HLEN;
	/* ETH_HLEN  14 */
	uip_len += ETH_HLEN;
	
	if (s < ETH_ZLEN) s = ETH_ZLEN;
	if (uip_len < ETH_ZLEN) uip_len = ETH_ZLEN;

	outb(D8390_COMMAND_PS0 | D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base+D8390_P0_COMMAND);
	outb(eth_tx_start, eth_nic_base+D8390_P0_TPSR);
	
	outb(s, eth_nic_base+D8390_P0_TBCR0);
	outb(uip_len, eth_nic_base+D8390_P0_TBCR0);
	
	outb(s>>8, eth_nic_base+D8390_P0_TBCR1);
	outb(uip_len>>8, eth_nic_base+D8390_P0_TBCR1);

  outb(D8390_COMMAND_PS0 | D8390_COMMAND_TXP | D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base+D8390_P0_COMMAND);
}
#endif

void ne2000_transmit(
	struct nic *nic,
	const char *d,			/* Destination */
	unsigned int t,			/* Type */
	unsigned int s,			/* size */
	const char *p)			/* Packet */
{
	/* Programmed I/O */
	unsigned short type;
	volatile char *tmpbuf;
	int i;

	/* 
	 * tmpbuf uip_buf[14] ּҸ ִ´. 
	 * 迭 14°  ̴   Application Ŷ ۵ȴ.
	 *   ּҸ ־ش.
	 */
	tmpbuf = &p[14];

	type = (t >> 8) | (t << 8);
	/*  MAC ּ */
	eth_pio_write(d, eth_tx_start<<8, ETH_ALEN);
	/* ߽ MAC ּ */
	eth_pio_write(nic->addr, (eth_tx_start<<8)+ETH_ALEN, ETH_ALEN);

	/* bcc generates worse code without (const+const) below */
	eth_pio_write((unsigned char *)&type, (eth_tx_start<<8)+(ETH_ALEN+ETH_ALEN), 2);
	/* eth_pio_write(p, (eth_tx_start<<8)+ETH_HLEN, s); */
	eth_pio_write(tmpbuf, (eth_tx_start<<8)+ETH_HLEN, s);
	
	s += ETH_HLEN;
	if (s < ETH_ZLEN) s = ETH_ZLEN;

	outb(D8390_COMMAND_PS0 | D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base+D8390_P0_COMMAND);
	outb(eth_tx_start, eth_nic_base+D8390_P0_TPSR);
	outb(s, eth_nic_base+D8390_P0_TBCR0);
	outb(s>>8, eth_nic_base+D8390_P0_TBCR1);
	outb(D8390_COMMAND_PS0 | D8390_COMMAND_TXP | D8390_COMMAND_RD2 | D8390_COMMAND_STA, eth_nic_base+D8390_P0_COMMAND);
}

/**************************************************************************
NS8390_POLL - Wait for a frame
**************************************************************************/
int ne2000_poll(struct nic *nic)
{
	int ret = 0;
	unsigned char rstat, curr, next;
	unsigned short len, frag;
	unsigned short pktoff;
	/* unsigned char *p; */
	struct ringbuffer pkthdr;

	rstat = inb(eth_nic_base+D8390_P0_RSR);
	if (!(rstat & D8390_RSTAT_PRX)) return(0);
	next = inb(eth_nic_base+D8390_P0_BOUND)+1;
	if (next >= eth_memsize) next = eth_rx_start;
	outb(D8390_COMMAND_PS1, eth_nic_base+D8390_P0_COMMAND);
	curr = inb(eth_nic_base+D8390_P1_CURR);
	outb(D8390_COMMAND_PS0, eth_nic_base+D8390_P0_COMMAND);
	if (curr >= eth_memsize) curr=eth_rx_start;
	if (curr == next) return(0);
	pktoff = next << 8;
	if (eth_flags & FLAG_PIO)
		eth_pio_read(pktoff, (char *)&pkthdr, 4);
	else
		memcpy(&pkthdr, (char *)eth_rmem + pktoff, 4);
	pktoff += sizeof(pkthdr);
	/* incoming length includes FCS so must sub 4 */
	len = pkthdr.len - 4;
	if ((pkthdr.status & D8390_RSTAT_PRX) == 0 || len < ETH_ZLEN
		|| len > ETH_FRAME_LEN) {
		/* printf("Bogus packet, ignoring Hey~\n"); */
		return (0);
	}
	else {
		/* p = nic->packet; */
    *uip_buf = nic->packet;
		nic->packetlen = len;		/* available to caller */
		frag = (eth_memsize << 8) - pktoff;
		if (len > frag) {		/* We have a wrap-around */
			/* read first part */
			if (eth_flags & FLAG_PIO)
				/* eth_pio_read(pktoff, p, frag); */
				eth_pio_read(pktoff, uip_buf, frag);
			else
				/* memcpy(p, (char *)eth_rmem + pktoff, frag); */
				memcpy(uip_buf, (char *)eth_rmem + pktoff, frag);
			pktoff = eth_rx_start << 8;
			/* p += frag; */
			*uip_buf += frag;
			len -= frag;
		}
		/* read second part */
		if (eth_flags & FLAG_PIO)
			/* eth_pio_read(pktoff, p, len); */
			eth_pio_read(pktoff, uip_buf, len);
		else
			/* memcpy(p, (char *)eth_rmem + pktoff, len); */
			memcpy(uip_buf, (char *)eth_rmem + pktoff, len);
		ret = 1;
	}
	next = pkthdr.next;		/* frame number of next packet */
	if (next == eth_rx_start)
		next = eth_memsize;
	outb(next-1, eth_nic_base+D8390_P0_BOUND);
	
	return(len + 4);
}
